@@ -13,9 +13,10 @@
 #![deny(rust_2018_idioms)]
 
 use base64::Engine;
-use pyo3::prelude::PyBytesMethods;
+use pyo3::types::PyBytesMethods;
 use pyo3::PyTypeInfo;
 use std::convert::TryInto;
+use std::ffi::CString;
 use std::io::Write;
 use subtle::ConstantTimeEq;
 
@@ -24,7 +25,8 @@ pub const BASE64_ENGINE: base64::engine::GeneralPurpose = base64::engine::Genera
     base64::engine::general_purpose::NO_PAD,
 );
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
+#[pyo3(signature = (rounds=None, prefix=None))]
 fn gensalt<'p>(
     py: pyo3::Python<'p>,
     rounds: Option<u16>,
@@ -48,7 +50,7 @@ fn gensalt<'p>(
 
     let encoded_salt = BASE64_ENGINE.encode(salt);
 
-    pyo3::types::PyBytes::new_bound_with(
+    pyo3::types::PyBytes::new_with(
         py,
         1 + prefix.len() + 1 + 2 + 1 + encoded_salt.len(),
         |mut b| {
@@ -64,7 +66,7 @@ fn gensalt<'p>(
     )
 }
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
 fn hashpw<'p>(
     py: pyo3::Python<'p>,
     password: &[u8],
@@ -113,13 +115,13 @@ fn hashpw<'p>(
     let hashed = py
         .allow_threads(|| bcrypt::hash_with_salt(password, cost, raw_salt))
         .map_err(|_| pyo3::exceptions::PyValueError::new_err("Invalid salt"))?;
-    Ok(pyo3::types::PyBytes::new_bound(
+    Ok(pyo3::types::PyBytes::new(
         py,
         hashed.format_for_version(version).as_bytes(),
     ))
 }
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
 fn checkpw(py: pyo3::Python<'_>, password: &[u8], hashed_password: &[u8]) -> pyo3::PyResult<bool> {
     Ok(hashpw(py, password, hashed_password)?
         .as_bytes()
@@ -127,17 +129,16 @@ fn checkpw(py: pyo3::Python<'_>, password: &[u8], hashed_password: &[u8]) -> pyo
         .into())
 }
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
+#[pyo3(signature = (password, salt, desired_key_bytes, rounds, ignore_few_rounds=false))]
 fn kdf<'p>(
     py: pyo3::Python<'p>,
     password: &[u8],
     salt: &[u8],
     desired_key_bytes: usize,
     rounds: u32,
-    ignore_few_rounds: Option<bool>,
+    ignore_few_rounds: bool,
 ) -> pyo3::PyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
-    let ignore_few_rounds = ignore_few_rounds.unwrap_or(false);
-
     if password.is_empty() || salt.is_empty() {
         return Err(pyo3::exceptions::PyValueError::new_err(
             "password and salt must not be empty",
@@ -160,15 +161,15 @@ fn kdf<'p>(
         // They probably think bcrypt.kdf()'s rounds parameter is logarithmic,
         // expecting this value to be slow enough (it probably would be if this
         // were bcrypt). Emit a warning.
-        pyo3::PyErr::warn_bound(
+        pyo3::PyErr::warn(
             py,
-            &pyo3::exceptions::PyUserWarning::type_object_bound(py),
-            &format!("Warning: bcrypt.kdf() called with only {rounds} round(s). This few is not secure: the parameter is linear, like PBKDF2."),
+            &pyo3::exceptions::PyUserWarning::type_object(py),
+            &CString::new(format!("Warning: bcrypt.kdf() called with only {rounds} round(s). This few is not secure: the parameter is linear, like PBKDF2.")).unwrap(),
             3
         )?;
     }
 
-    pyo3::types::PyBytes::new_bound_with(py, desired_key_bytes, |output| {
+    pyo3::types::PyBytes::new_with(py, desired_key_bytes, |output| {
         py.allow_threads(|| {
             bcrypt_pbkdf::bcrypt_pbkdf(password, salt, rounds, output).unwrap();
         });
@@ -176,34 +177,35 @@ fn kdf<'p>(
     })
 }
 
-#[pyo3::prelude::pymodule]
-fn _bcrypt(
-    _py: pyo3::Python<'_>,
-    m: &pyo3::Bound<'_, pyo3::types::PyModule>,
-) -> pyo3::PyResult<()> {
-    m.add_function(pyo3::wrap_pyfunction!(gensalt, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(hashpw, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(checkpw, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(kdf, m)?)?;
+#[pyo3::pymodule]
+mod _bcrypt {
+    use pyo3::types::PyModuleMethods;
 
-    m.add("__title__", "bcrypt")?;
-    m.add(
-        "__summary__",
-        "Modern(-ish) password hashing for your software and your servers",
-    )?;
-    m.add("__uri__", "https://github.com/pyca/bcrypt/")?;
+    #[pymodule_export]
+    use super::{checkpw, gensalt, hashpw, kdf};
 
-    // When updating this, also update pyproject.toml
-    // This isn't named __version__ because passlib treats the existence of
-    // that attribute as proof that we're a different module
-    m.add("__version_ex__", "4.1.3")?;
+    // Not yet possible to add constants declaratively.
+    #[pymodule_init]
+    fn init(m: &pyo3::Bound<'_, pyo3::types::PyModule>) -> pyo3::PyResult<()> {
+        m.add("__title__", "bcrypt")?;
+        m.add(
+            "__summary__",
+            "Modern(-ish) password hashing for your software and your servers",
+        )?;
+        m.add("__uri__", "https://github.com/pyca/bcrypt/")?;
 
-    let author = "The Python Cryptographic Authority developers";
-    m.add("__author__", author)?;
-    m.add("__email__", "cryptography-dev@python.org")?;
+        // When updating this, also update pyproject.toml
+        // This isn't named __version__ because passlib treats the existence of
+        // that attribute as proof that we're a different module
+        m.add("__version_ex__", "4.2.1")?;
 
-    m.add("__license__", "Apache License, Version 2.0")?;
-    m.add("__copyright__", format!("Copyright 2013-2024 {author}"))?;
+        let author = "The Python Cryptographic Authority developers";
+        m.add("__author__", author)?;
+        m.add("__email__", "cryptography-dev@python.org")?;
 
-    Ok(())
+        m.add("__license__", "Apache License, Version 2.0")?;
+        m.add("__copyright__", format!("Copyright 2013-2024 {author}"))?;
+
+        Ok(())
+    }
 }
